__all__ = ["app", "cli"]

# --- zyx ----------------------------------------------------------------

from textual.app import App, ComposeResult
from textual.widgets import Input, RichLog
from textual.containers import VerticalScroll
from typing import Any, Union, Optional, List, Callable, Literal
from ..types import ClientModeParams
from ..core.ext import BaseModel

COLOR_MAP = {
    "black": "#000000",
    "white": "#FFFFFF",
    "red": "#FF0000",
    "green": "#008000",
    "blue": "#0000FF",
    "yellow": "#FFFF00",
    "cyan": "#00FFFF",
    "magenta": "#FF00FF",
    "silver": "#C0C0C0",
    "deep_blue": "#001f3f",
    "ocean_blue": "#0074D9",
    "sunset_orange": "#FF851B",
    "twilight_purple": "#6F42C1",
    "forest_green": "#2ECC40",
    "midnight_black": "#111111",
    "crimson_red": "#DC143C",
    "royal_gold": "#FFD700",
    "peach": "#FFDAB9",
    "lavender": "#E6E6FA",
    "teal": "#008080",
    "coral": "#FF7F50",
    "mustard_yellow": "#FFDB58",
    "powder_blue": "#B0E0E6",
    "sage_green": "#B2AC88",
    "blush": "#FF6F61",
    "steel_grey": "#7A8B8B",
    "ice_blue": "#AFEEEE",
    "burnt_sienna": "#E97451",
    "plum": "#DDA0DD",
    "emerald_green": "#50C878",
    "ruby_red": "#E0115F",
    "sapphire_blue": "#0F52BA",
    "amethyst_purple": "#9966CC",
    "topaz_yellow": "#FFC87C",
    "turquoise": "#40E0D0",
    "rose_gold": "#B76E79",
    "olive_green": "#808000",
    "burgundy": "#800020",
    "navy_blue": "#000080",
    "mauve": "#E0B0FF",
    "chartreuse": "#7FFF00",
    "terracotta": "#E2725B",
    "indigo": "#4B0082",
    "periwinkle": "#CCCCFF",
    "maroon": "#800000",
    "cerulean": "#007BA7",
    "ochre": "#CC7722",
    "slate_gray": "#708090",
    "mint_green": "#98FF98",
    "salmon": "#FA8072",
    "tangerine": "#F28500",
    "taupe": "#483C32",
    "aquamarine": "#7FFFD4",
    "mahogany": "#C04000",
    "fuchsia": "#FF00FF",
    "azure": "#007FFF",
    "lilac": "#C8A2C8",
    "vermilion": "#E34234",
    "ivory": "#FFFFF0",
}

ColorName = Literal[
    "black",
    "white",
    "red",
    "green",
    "blue",
    "yellow",
    "cyan",
    "magenta",
    "silver",
    "deep_blue",
    "ocean_blue",
    "sunset_orange",
    "twilight_purple",
    "forest_green",
    "midnight_black",
    "crimson_red",
    "royal_gold",
    "peach",
    "lavender",
    "teal",
    "coral",
    "mustard_yellow",
    "powder_blue",
    "sage_green",
    "blush",
    "steel_grey",
    "ice_blue",
    "burnt_sienna",
    "plum",
    "emerald_green",
    "ruby_red",
    "sapphire_blue",
    "amethyst_purple",
    "topaz_yellow",
    "turquoise",
    "rose_gold",
    "olive_green",
    "burgundy",
    "navy_blue",
    "mauve",
    "chartreuse",
    "terracotta",
    "indigo",
    "periwinkle",
    "maroon",
    "cerulean",
    "ochre",
    "slate_gray",
    "mint_green",
    "salmon",
    "tangerine",
    "taupe",
    "aquamarine",
    "mahogany",
    "fuchsia",
    "azure",
    "lilac",
    "vermilion",
    "ivory",
]


class ChatApp(App):
    def __init__(
        self,
        provider: Optional[Any] = None,
        messages: Union[str, list[dict]] = None,
        model: Optional[str] = "gpt-4o-mini",
        tools: Optional[List[Union[Callable, dict, BaseModel]]] = None,
        run_tools: Optional[bool] = True,
        response_model: Optional[BaseModel] = None,
        mode: Optional[ClientModeParams] = "tools",
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        top_p: Optional[float] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        max_retries: Optional[int] = 3,
        verbose: Optional[bool] = False,
        background: Optional[Union[str, ColorName]] = "midnight_black",
        text: Optional[Union[str, ColorName]] = "steel_grey",
        input_field: Optional[Union[str, ColorName]] = "ocean_blue",
        **kwargs,
    ):
        try:
            from .main import Client

            super().__init__()
            self.client = Client()
            self.provider = provider
            self.chat_history = (
                self.client.format_messages(messages) if messages else []
            )
            self.model = model
            self.tools = tools
            self.run_tools = run_tools
            self.response_model = response_model
            self.mode = mode
            self.base_url = base_url
            self.api_key = api_key
            self.organization = organization
            self.top_p = top_p
            self.temperature = temperature
            self.max_tokens = max_tokens
            self.max_retries = max_retries
            self.verbose = verbose
            self.background = COLOR_MAP.get(background, background)
            self.text = COLOR_MAP.get(text, text)
            self.input_field = COLOR_MAP.get(input_field, input_field)
            self.kwargs = kwargs
        except Exception as e:
            print(f"Error initializing ChatApp: {e}")

    def compose(self) -> ComposeResult:
        try:
            self.CSS = f"""
            ChatApp {{
                background: {self.background};
            }}
            
            RichLog#chat_display {{
                border: round $primary;
                background: #2e2e2e;
                color: {self.text};
                padding: 1 2;
            }}
        
            Input#input_field {{
                border: round $primary;
                background: {self.input_field};
                color: $text;
                padding: 1 2;
            }}
            """

            with VerticalScroll():
                yield RichLog(id="chat_display")
                yield Input(placeholder="Type your message...", id="input_field")
        except Exception as e:
            print(f"Error composing ChatApp: {e}")

    def on_input_submitted(self, event: Input.Submitted) -> None:
        try:
            from rich.text import Text

            user_message = event.value.strip()
            if user_message:
                self.chat_history.append({"role": "user", "content": user_message})

                user_text = Text()
                user_text.append("User: ", style="bold blue")
                user_text.append(f"{user_message}\n")

                self.query_one("#chat_display", RichLog).write(user_text)

                if self.provider:
                    # Use the provider (Agents instance) to handle the message
                    response = self.provider.run(user_message, **self.kwargs)
                else:
                    # Use the completion client directly
                    response = self.client.completion(
                        messages=self.chat_history,
                        model=self.model,
                        tools=self.tools,
                        run_tools=self.run_tools,
                        response_model=self.response_model,
                        mode=self.mode,
                        base_url=self.base_url,
                        api_key=self.api_key,
                        organization=self.organization,
                        top_p=self.top_p,
                        temperature=self.temperature,
                        max_tokens=self.max_tokens,
                        max_retries=self.max_retries,
                        verbose=self.verbose,
                        **self.kwargs,
                    )
                    assistant_reply = response.choices[0].message["content"]
                    self.chat_history.append(
                        {"role": "assistant", "content": assistant_reply}
                    )
                    response = assistant_reply

                assistant_text = Text()
                assistant_text.append("Assistant: ", style="bold green")
                assistant_text.append(f"{response}\n")

                self.query_one("#chat_display", RichLog).write(assistant_text)

                self.query_one("#input_field", Input).value = ""
        except Exception as e:
            print(f"Error processing input: {e}")


def cli(
    provider: Optional[Any] = None,
    messages: Union[str, list[dict]] = None,
    model: Optional[str] = "gpt-4o-mini",
    tools: Optional[List[Union[Callable, dict, BaseModel]]] = None,
    run_tools: Optional[bool] = True,
    response_model: Optional[BaseModel] = None,
    mode: Optional[ClientModeParams] = "tools",
    base_url: Optional[str] = None,
    api_key: Optional[str] = None,
    organization: Optional[str] = None,
    top_p: Optional[float] = None,
    temperature: Optional[float] = None,
    max_tokens: Optional[int] = None,
    max_retries: Optional[int] = 3,
    verbose: Optional[bool] = False,
    background: Optional[Union[str, ColorName]] = "midnight_black",
    text: Optional[Union[str, ColorName]] = "steel_grey",
    input_field: Optional[Union[str, ColorName]] = "ocean_blue",
    **kwargs,
):
    """Runs an easy to use CLI interface for a Chatbot, using either an Agents instance or the .completion() function.

    Parameters:
        - provider (Optional[Any]): The provider instance (e.g., Agents) to handle the chat logic.
        - messages (Union[str, list[dict]]): The messages to send to the model.
        - model (Optional[str]): The model to use for completions.
        - tools (Optional[List[Union[Callable, dict, BaseModel]]]): The tools to use for completions.
        - run_tools (Optional[bool]): Whether to run the tools.
        - response_model (Optional[BaseModel]): The Pydantic response model to use for completions.
        - background (Optional[Union[str, ColorName]]): Background color of the app.
        - text (Optional[Union[str, ColorName]]): Text color in the chat display.
        - input_field (Optional[Union[str, ColorName]]): Color of the input field.
        - mode (Optional[str]): The mode to use for completions.
        - base_url (Optional[str]): The base URL for the API.
        - api_key (Optional[str]): The API key to use for the API.
        - organization (Optional[str]): The organization to use for the API.
        - top_p (Optional[float]): The top-p value for completions.
        - temperature (Optional[float]): The temperature value for completions.
        - max_tokens (Optional[int]): The maximum number of tokens for completions.
        - max_retries (Optional[int]): The maximum number of retries for completions.
        - verbose (Optional[bool]): Whether to print verbose output.
    """

    try:
        ChatApp(
            provider=provider,
            messages=messages,
            model=model,
            tools=tools,
            run_tools=run_tools,
            response_model=response_model,
            mode=mode,
            base_url=base_url,
            api_key=api_key,
            organization=organization,
            top_p=top_p,
            temperature=temperature,
            max_tokens=max_tokens,
            max_retries=max_retries,
            verbose=verbose,
            background=background,
            text=text,
            input_field=input_field,
            **kwargs,
        ).run()
    except Exception as e:
        print(f"Error running ChatApp: {e}")


# --- zyx ----------------------------------------------------------------


class CustomChatApp(App):
    def __init__(
        self,
        input_handler: Callable[[str], Any],
        output_handler: Callable[[Any], str],
        background: Optional[Union[str, ColorName]] = "midnight_black",
        text: Optional[Union[str, ColorName]] = "steel_grey",
        input_field: Optional[Union[str, ColorName]] = "ocean_blue",
    ):
        try:
            super().__init__()
            self.input_handler = input_handler
            self.output_handler = output_handler
            self.background = COLOR_MAP.get(background, background)
            self.text = COLOR_MAP.get(text, text)
            self.input_field = COLOR_MAP.get(input_field, input_field)
        except Exception as e:
            print(f"Error initializing CustomChatApp: {e}")

    def compose(self) -> ComposeResult:
        try:
            self.CSS = f"""
            CustomChatApp {{
                background: {self.background};
            }}
            
            RichLog#chat_display {{
                border: round $primary;
                background: #2e2e2e;
                color: {self.text};
                padding: 1 2;
            }}
        
            Input#input_field {{
                border: round $primary;
                background: {self.input_field};
                color: $text;
                padding: 1 2;
            }}
            """
            with VerticalScroll():
                yield RichLog(id="chat_display")
                yield Input(placeholder="Type your message...", id="input_field")
        except Exception as e:
            print(f"Error composing CustomChatApp: {e}")

    def on_input_submitted(self, event: Input.Submitted) -> None:
        try:
            from rich.text import Text

            user_message = event.value.strip()
            if user_message:
                user_text = Text()
                user_text.append("User: ", style="bold blue")
                user_text.append(f"{user_message}\n")
                self.query_one("#chat_display", RichLog).write(user_text)

                processed_input = self.input_handler(user_message)

                response = self.output_handler(processed_input)

                assistant_text = Text()
                assistant_text.append("Assistant: ", style="bold green")
                assistant_text.append(f"{response}\n")
                self.query_one("#chat_display", RichLog).write(assistant_text)

                self.query_one("#input_field", Input).value = ""
        except Exception as e:
            print(f"Error processing input in CustomChatApp: {e}")


def app(
    input_handler: Callable[[str], Any],
    output_handler: Callable[[Any], str],
    background: Optional[Union[str, ColorName]] = "midnight_black",
    text: Optional[Union[str, ColorName]] = "steel_grey",
    input_field: Optional[Union[str, ColorName]] = "ocean_blue",
):
    """
    Creates and runs a custom chat application with user-defined input and output handlers.

    Parameters:
    - input_handler: A function that takes a string (user input) and returns any type of data.
    - output_handler: A function that takes the result of input_handler and returns a string (assistant's response).
    - background: Background color of the app.
    - text: Text color in the chat display.
    - input_field: Color of the input field.

    Example usage:
    def custom_input_handler(user_input: str) -> dict:
        return {"user_message": user_input}

    def custom_output_handler(processed_input: dict) -> str:
        return f"Received: {processed_input['user_message']}"

    app(custom_input_handler, custom_output_handler)
    """
    try:
        CustomChatApp(
            input_handler=input_handler,
            output_handler=output_handler,
            background=background,
            text=text,
            input_field=input_field,
        ).run()
    except Exception as e:
        print(f"Error running CustomChatApp: {e}")



from qdrant_client.http import models as rest
from pydantic import BaseModel, Field
from typing import Callable, List, Union, Optional, Literal
import uuid
import logging

from ..types import (
    ClientModeParams,
    ClientParams
)
from .. import logger

class QdrantNode(BaseModel):
    id: str
    text: str
    embedding: List[float]
    metadata: Optional[dict] = None

class SearchResponse(BaseModel):
    query: str
    results: List[QdrantNode] = Field(default_factory=list)

class Qdrant:
    def __init__(
        self,
        collection_name: str = "my_collection",
        vector_size: int = 1536,
        distance: Literal["Cosine", "Euclid", "Dot"] = "Cosine",
        location : str = ":memory:",
        host : str = None,
        port: int = 6333,
        embedding_model: str = "text-embedding-3-small",
        embedding_dimensions: Optional[int] = None,
        embedding_api_key: Optional[str] = None,
        embedding_api_base: Optional[str] = None,
        embedding_api_version: Optional[str] = None
    ):
        from qdrant_client.http.models import Distance
        from qdrant_client import QdrantClient
        
        self.client = QdrantClient(location = location, host = host, port=port)
        self.collection_name = collection_name
        self.vector_size = vector_size
        self.distance = Distance(distance)
        self.embedding_model = embedding_model
        self.embedding_dimensions = embedding_dimensions
        self.embedding_api_key = embedding_api_key
        self.embedding_api_base = embedding_api_base
        self.embedding_api_version = embedding_api_version

        self._create_collection()

    def _create_collection(self):
        from qdrant_client.http.models import VectorParams
        try:
            self.client.get_collection(self.collection_name)
            logger.info(f"Collection '{self.collection_name}' already exists.")
        except Exception as e:
    
            logger.info(f"Collection '{self.collection_name}' does not exist. Creating it now.")
            self.client.create_collection(
                collection_name=self.collection_name,
                vectors_config=VectorParams(size=self.vector_size, distance=self.distance)
            )
            logger.info(f"Collection '{self.collection_name}' created successfully.")

    def _get_embedding(self, text: str) -> List[float]:
        from litellm.main import embedding as litellm_embedding
        
        try:
            response = litellm_embedding(
                model=self.embedding_model,
                input=[text],
                dimensions=self.embedding_dimensions,
                api_key=self.embedding_api_key,
                api_base=self.embedding_api_base,
                api_version=self.embedding_api_version
            )

            embedding_data = response.get('data', None)
            if embedding_data and isinstance(embedding_data, list) and len(embedding_data) > 0:
            
                embedding_vector = embedding_data[0].get('embedding', None)
                if isinstance(embedding_vector, list) and all(isinstance(x, float) for x in embedding_vector):
                    return embedding_vector
                else:
                    raise ValueError("Invalid embedding format: Expected a list of floats within the 'embedding' key")
            else:
                raise ValueError("Embedding data is missing or improperly formatted")
        except Exception as e:
            logger.error(f"Error generating embedding: {e}")
            raise

    def add(self, texts: Union[str, List[str]], metadata: Optional[dict] = None):
        from qdrant_client.http.models import PointStruct
        
        if isinstance(texts, str):
            texts = [texts]

        points = []
        for text in texts:
            try:
                embedding_vector = self._get_embedding(text)
                node_id = str(uuid.uuid4())
                point = PointStruct(
                    id=str(uuid.uuid4()),
                    vector=embedding_vector,
                    payload={'text': text, 'metadata': metadata or {}}
                )
                points.append(point)
            except Exception as e:
                logger.error(f"Error processing text: {text}. Error: {e}")

        if points:
            try:
                self.client.upsert(
                    collection_name=self.collection_name,
                    points=[point]
                )
                logger.info(f"Successfully added {len(points)} points to the collection.")
            except Exception as e:
                logger.error(f"Error upserting points to collection: {e}")
        else:
            logger.warning("No valid points to add to the collection.")

    def search(self, query: str, top_k: int = 5) -> SearchResponse:
        try:
            query_vector = self._get_embedding(query)
            search_results = self.client.search(
                collection_name=self.collection_name,
                query_vector=query_vector,
                limit=top_k
            )
            
            nodes = []
            for result in search_results:
                payload = result.payload
                node = QdrantNode(
                    id=str(result.id),
                    text=payload.get('text', ''),
                    embedding=query_vector, 
                    metadata=payload.get('metadata', {})
                )
                nodes.append(node)
            return SearchResponse(query=query, results=nodes)
        except Exception as e:
            logger.error(f"Error during search: {e}")
            raise
        
    def completion(
        self,
        messages : Union[str, list[dict[str, str]]],
        model : Optional[str] = "gpt-4o-mini",
        top_k : Optional[int] = 5,
        tools: Optional[List[Union[Callable, dict, BaseModel]]] = None,
        run_tools: Optional[bool] = True,
        response_model: Optional[BaseModel] = None,
        mode: Optional[ClientModeParams] = "tools",
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        top_p: Optional[float] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        max_retries: Optional[int] = 3,
        verbose: Optional[bool] = False,
    ):
        if isinstance(messages, str):
            query = messages
            messages = [{"role" : "user", "content" : messages}]
        elif isinstance(messages, list):
            query = messages[-1].get("content", "")
            
        try:
            results = self.search(query, top_k = top_k)
        except Exception as e:
            logger.error(f"Error during search: {e}")
            results = SearchResponse(query=query)
            
        if messages and isinstance(messages, list):
            if not any(message.get("role", "") == "system" for message in messages):
                messages.insert(0, {"role" : "system", "content" : f"You have retrieved the following relevant information. Use only if relevant {results}"})
            else:
                for message in messages:
                    if message.get("role", "") == "system":
                        message["content"] += f" You have retrieved the following relevant information. Use only if relevant {results}"
                        break
        from .main import Client
        return Client().completion(
            messages = messages,
            model = model,
            tools = tools,
            run_tools = run_tools,
            response_model = response_model,
            mode = mode,
            base_url = base_url,
            api_key = api_key,
            organization = organization,
            top_p = top_p,
            temperature = temperature,
            max_tokens = max_tokens,
            max_retries = max_retries,
            verbose = verbose,
        )
        

if __name__ == '__main__':
    try:
        qdrant = Qdrant(collection_name="my_collection", vector_size=1536)
        qdrant.add(["Hello, world!", "How are you?", "What's up?"])
        results = qdrant.search("How are you?")
        for result in results.results:
            print(f"ID: {result.id}, Text: {result.text}, Metadata: {result.metadata}")
    except Exception as e:
        logger.error(f"Error in main execution: {e}")



__all__ = ["Client", "completion"]

# --- zyx ----------------------------------------------------------------

from ..types import ClientParams, ClientModeParams, ClientToolParams
from ..utils.tool_calling import convert_to_openai_tool
from ..core.ext import BaseModel
from typing import Callable, List, Optional, Union, Generator
from litellm.types.utils import ModelResponse

from litellm import embedding
class Client:
    def __init__(
        self,
    ):
        pass

    @staticmethod
    def format_messages(messages: Union[str, list[dict]] = None) -> list[dict]:
        try:
            if isinstance(messages, str):
                return [{"role": "user", "content": messages}]
            elif isinstance(messages, list) and all(isinstance(m, dict) for m in messages):
                return messages
            else:
                raise ValueError("Invalid message format")
        except Exception as e:
            print(f"Error formatting messages: {e}")
            return []

    @staticmethod
    def format_tools(tools: List[Union[Callable, dict, BaseModel]] = None):
        try:
            if not tools:
                return None
            params = ClientToolParams(tools=tools, openai_tools=[], mapping={})
            for tool in tools:
                if isinstance(tool, dict):
                    params.openai_tools.append(tool)
                elif isinstance(tool, BaseModel) or isinstance(tool, Callable):
                    params.openai_tools.append(convert_to_openai_tool(tool))
            for tool in tools:
                if isinstance(tool, Callable) or isinstance(tool, BaseModel):
                    params.mapping[tool.__name__] = tool
            return params
        except Exception as e:
            print(f"Error formatting tools: {e}")
            return None

    def run_base_completion(self):
        try:
            from litellm.main import completion

            return completion(
                model=self.params.model,
                messages=self.params.messages,
                base_url=self.params.base_url,
                tools=self.params.tools.openai_tools if self.params.tools else None,
                api_key=self.params.api_key,
                organization=self.params.organization,
                top_p=self.params.top_p,
                temperature=self.params.temperature,
                max_tokens=self.params.max_tokens,
                **self.params.kwargs,
            )
        except Exception as e:
            print(f"Error running base completion: {e}")
            return None
        
    def stream_completion(self):
        try:
            from litellm.main import completion

            response = completion(
                model=self.params.model,
                messages=self.params.messages,
                base_url=self.params.base_url,
                tools=self.params.tools.openai_tools if self.params.tools else None,
                api_key=self.params.api_key,
                organization=self.params.organization,
                top_p=self.params.top_p,
                temperature=self.params.temperature,
                max_tokens=self.params.max_tokens,
                stream=True,
                **self.params.kwargs,
            )

            for chunk in response:
                yield chunk.choices[0].delta.content or ""

        except Exception as e:
            print(f"Error in stream_completion: {e}")
            yield None

    def run_instructor_completion(self):
        try:
            from litellm.main import completion
            from instructor import from_litellm, Mode

            if self.params.mode == "json":
                mode = Mode.JSON
            elif self.params.mode == "json_schema":
                mode = Mode.JSON_SCHEMA
            elif self.params.mode == "md_json":
                mode = Mode.MD_JSON
            elif self.params.mode == "parallel":
                mode = Mode.PARALLEL_TOOLS
            elif self.params.mode == "tools":
                mode = Mode.TOOLS
            client = from_litellm(completion, mode=mode)
            return client.chat.completions.create(
                messages=self.params.messages,
                model=self.params.model,
                response_model=self.params.response_model,
                base_url=self.params.base_url,
                api_key=self.params.api_key,
                tools=self.params.tools.openai_tools if self.params.tools else None,
                organization=self.params.organization,
                top_p=self.params.top_p,
                temperature=self.params.temperature,
                max_tokens=self.params.max_tokens,
                max_retries=self.params.max_retries,
                **self.params.kwargs,
            )
        except Exception as e:
            print(f"Error running instructor completion: {e}")
            return None

    def execute_tools(self, tool_calls: list[dict]):
        try:
            import json
            for tool_call in tool_calls:
                function_name = tool_call.function.name
                if function_name not in self.params.tools.mapping:
                    self.params.messages.append({
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": function_name,
                        "content": "Tool not executable",
                    })
                    continue
                function_args = json.loads(tool_call.function.arguments)
                tool_result = self.params.tools.mapping[function_name](**function_args)
                self.params.messages.append({
                    "tool_call_id": tool_call.id,
                    "role": "tool",
                    "name": function_name,
                    "content": str(tool_result),
                })
            
            return self.params.messages
        except Exception as e:
            print(f"Error executing tools: {e}")
            return self.params.messages

    def completion(
        self,
        messages: Union[str, list[dict]] = None,
        model: Optional[str] = "gpt-4o-mini",
        tools: Optional[List[Union[Callable, dict, BaseModel]]] = None,
        run_tools: Optional[bool] = True,
        response_model: Optional[BaseModel] = None,
        mode: Optional[ClientModeParams] = "tools",
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        top_p: Optional[float] = None,
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = None,
        max_retries: Optional[int] = 3,
        params: Optional[ClientParams] = None,
        verbose: Optional[bool] = False,
        **kwargs,
    ):
        """Runs a tool calling & structured output capable completion or completion chain.

        Example:
            ```python

            import zyx

            zyx.completion()
            ```

        Parameters:
            - messages (Union[str, list[dict]]): The messages to send to the model.
            - model (Optional[str]): The model to use for completion.
            - tools (Optional[List[Union[Callable, dict, BaseModel]]]): The tools to use for completion.
            - run_tools (Optional[bool]): Whether to run the tools.
            - response_model (Optional[BaseModel]): The response model to use for completion.
            - mode (Optional[ClientModeParams]): The mode to use for completion.
            - base_url (Optional[str]): The base URL to use for completion.
            - api_key (Optional[str]): The API key to use for completion.
            - organization (Optional[str]): The organization to use for completion.
            - top_p (Optional[float]): The top p value to use for completion.
            - temperature (Optional[float]): The temperature value to use for completion.
            - max_tokens (Optional[int]): The maximum tokens to use for completion.
            - max_retries (Optional[int]): The maximum retries to use for completion.
            - params (Optional[ClientParams]): The parameters to use for completion.
            - verbose (Optional[bool]): Whether to print the parameters.

        Returns:
            - The completion response.
        """
        try:
            if params:
                self.params = params
            else:
                self.params = ClientParams(
                    messages=self.format_messages(messages),
                    model=model,
                    tools=self.format_tools(tools),
                    run_tools=run_tools,
                    response_model=response_model,
                    mode=mode,
                    base_url=base_url,
                    api_key=api_key,
                    organization=organization,
                    top_p=top_p,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    max_retries=max_retries,
                    kwargs=kwargs,
                )
            if verbose:
                print(self.params + "\n")

            if not self.params.tools or not self.params.run_tools:
                if not self.params.response_model:
                    return self.run_base_completion()
                else:
                    return self.run_instructor_completion()
            response = self.run_base_completion()

            if not self.params.response_model:
                if not response.choices[0].message.tool_calls:
                    return response
                else:
                    while response.choices[0].message.tool_calls:
                        message = response.choices[0].message
                        self.params.messages.append(message)
                        self.params.messages = self.execute_tools(
                            message.tool_calls
                        )
                        response = self.run_base_completion()
                    return response
            else:
                if not response.choices[0].message.tool_calls:
                    self.params.messages.append(response.choices[0].message)
                    self.params.messages.append(
                        {
                            "role": "user",
                            "content": "Append tool responses to BaseModel.",
                        }
                    )
                    self.params.tools = None
                    return self.run_instructor_completion()
                else:
                    while response.choices[0].message.tool_calls:
                        self.params.messages.append(response.choices[0].message)
                        self.params.messages = self.execute_tools(
                            response.choices[0].message.tool_calls
                        )
                        response = self.run_base_completion()
                    self.params.messages.append(
                        {
                            "role": "user",
                            "content": "Append tool responses to BaseModel.",
                        }
                    )
                    self.params.tools = None
                    return self.run_instructor_completion()
        except Exception as e:
            print(f"Error in completion: {e}")
            return None


def completion(
    messages: Union[str, list[dict]] = None,
    model: Optional[str] = "gpt-4o-mini",
    tools: Optional[List[Union[Callable, dict, BaseModel]]] = None,
    run_tools: Optional[bool] = True,
    response_model: Optional[BaseModel] = None,
    mode: Optional[ClientModeParams] = "tools",
    base_url: Optional[str] = None,
    api_key: Optional[str] = None,
    organization: Optional[str] = None,
    top_p: Optional[float] = None,
    temperature: Optional[float] = None,
    max_tokens: Optional[int] = None,
    max_retries: Optional[int] = 3,
    params: Optional[ClientParams] = None,
    verbose: Optional[bool] = False,
    stream: Optional[bool] = False,
    **kwargs,
) -> Union[ModelResponse, BaseModel, Generator]:
    """Runs a tool calling & structured output capable completion or completion chain.

    Example:
        ```python
        import zyx

        zyx.completion()
        ```

    Parameters:
        - messages (Union[str, list[dict]]): The messages to send to the model.
        - model (Optional[str]): The model to use for completion.
        - tools (Optional[List[Union[Callable, dict, BaseModel]]]): The tools to use for completion.
        - run_tools (Optional[bool]): Whether to run the tools.
        - response_model (Optional[BaseModel]): The response model to use for completion.
        - mode (Optional[ClientModeParams]): The mode to use for completion.
        - base_url (Optional[str]): The base URL to use for completion.
        - api_key (Optional[str]): The API key to use for completion.
        - organization (Optional[str]): The organization to use for completion.
        - top_p (Optional[float]): The top p value to use for completion.
        - temperature (Optional[float]): The temperature value to use for completion.
        - max_tokens (Optional[int]): The maximum tokens to use for completion.
        - max_retries (Optional[int]): The maximum retries to use for completion.
        - params (Optional[ClientParams]): The parameters to use for completion.
        - verbose (Optional[bool]): Whether to print the parameters.
        - stream (Optional[bool]): Whether to stream the response.

    Returns:
        - The completion response, or a generator if streaming.
    """
    try:
        client = Client()
        client_params = params or ClientParams(
            messages=client.format_messages(messages),
            model=model,
            tools=client.format_tools(tools),
            run_tools=run_tools,
            response_model=response_model,
            mode=mode,
            base_url=base_url,
            api_key=api_key,
            organization=organization,
            top_p=top_p,
            temperature=temperature,
            max_tokens=max_tokens,
            max_retries=max_retries,
            kwargs=kwargs,
        )
        client.params = client_params

        if verbose:
            print(client.params + "\n")

        if stream and not response_model:
            return client.stream_completion()
        else:
            return client.completion(
                messages=messages,
                model=model,
                tools=tools,
                run_tools=run_tools,
                response_model=response_model,
                mode=mode,
                base_url=base_url,
                api_key=api_key,
                organization=organization,
                top_p=top_p,
                temperature=temperature,
                max_tokens=max_tokens,
                max_retries=max_retries,
                params=params,
                verbose=verbose,
                **kwargs,
            )
    except Exception as e:
        print(f"Error in completion function: {e}")
        return None


__all__ = ["ClientModeParams", "ClientParams"]

# --- zyx ----------------------------------------------------------------

from ..core.ext import BaseModel, Field
from typing import Any, Callable, List, Literal, Optional, Union, Dict
from uuid import uuid4

# --- LLM Completion Client ---

ClientModeParams = Literal["json", "json_schema", "md_json", "parallel", "tools"]


class ClientToolParams(BaseModel):
    tools: List[Union[Callable, dict, BaseModel]] = None
    openai_tools: List[dict] = None
    mapping: Optional[dict] = None


class ClientParams(BaseModel):
    messages: Union[str, list[dict]] = None

    model: Optional[str] = "gpt-4o-mini"
    tools: Optional[ClientToolParams] = None
    run_tools: Optional[bool] = True
    response_model: Optional[Union[Any, BaseModel]] = None
    mode: Optional[ClientModeParams] = "tools"

    base_url: Optional[str] = None
    api_key: Optional[str] = None
    organization: Optional[str] = None

    top_p: Optional[float] = None
    temperature: Optional[float] = None
    max_tokens: Optional[int] = 3
    max_retries: Optional[int] = None
    kwargs: Optional[dict] = None


# --- Agents ---


class TaskIntent(BaseModel):
    intent: str
    description: str
    priority: int = Field(default=1, ge=1, le=5)


class TaskDelegation(BaseModel):
    task_id: str = Field(default_factory=lambda: str(uuid4()))
    intent: TaskIntent
    assigned_worker: Optional[str] = None
    status: str = "pending"
    result: Optional[Any] = None


class SupervisorResponse(BaseModel):
    message: Optional[str] = None
    delegations: Optional[List[TaskDelegation]] = None


class WorkerResponse(BaseModel):
    task_id: str
    status: str
    result: Any
    message: Optional[str] = None


class AgentParams(BaseModel):
    agent_id: str
    agent_type: str
    tools: List[Any] = []
    instructions: Optional[str] = None
    completion_params: Optional[ClientParams] = None
